// --------------------------------------------------------------------------------------------------------------------
// <copyright file="Sample.cs" company="Comtrade d.o.o.">
//   
// </copyright>
// <summary>
//   C# sample for FURS SOAP portal
// </summary>
// --------------------------------------------------------------------------------------------------------------------

namespace SoapPortalSample
{
  using System;
  using System.IO;
  using System.Security.Cryptography;
  using System.Security.Cryptography.X509Certificates;
  using System.Security.Cryptography.Xml;
  using System.ServiceModel;
  using System.Text;
  using System.Xml;
  using System.Xml.XPath;
  using System.Xml.Xsl;

  using SoapPortalSample.Service;

  /// <summary>
  /// C# sample for FURS SOAP portal
  /// </summary>
  internal class Sample
  {
    #region Constants and Fields

    public const string DmvoUri = "http://edavki.durs.si/Documents/Schemas/DMV_O_4.xsd";

    public const string DocumentInfoUri = "http://edavki.durs.si/WebServices/SoapPortal/Schemas/DocumentInfo_1.xsd";

    public const string HslsigUri = "urn:schemas-hermes-softlab-com:2003/09/Signatures";

    public const string NfPrevZavUri = "http://edavki.durs.si/Documents/Schemas/NF_PrevZav_1.xsd";

    private const string DmvoTemplate =
      @"<?xml version=""1.0"" encoding=""utf-8""?>
<Envelope xmlns=""http://edavki.durs.si/Documents/Schemas/DMV_O_4.xsd"" xmlns:edp=""http://edavki.durs.si/Documents/Schemas/EDP-Common-1.xsd"">
  <edp:Header>
  </edp:Header>
  <edp:AttachmentList />
  <edp:Signatures />
  <body>
    <edp:bodyContent />
    <DMV_O>
      <DocumentWorkflowID>O</DocumentWorkflowID>
      <Period>201503</Period>
      <TotalTax>27300</TotalTax>
    </DMV_O>
  </body>
</Envelope>";

    private const string DsigUri = "http://www.w3.org/2000/09/xmldsig#";

    private const string EdpUri = "http://edavki.durs.si/Documents/Schemas/EDP-Common-1.xsd";

    private static string certificateName = "Test1";

    private static string documentDomain = "beta.edavki.durs.si";

    private static string soapServiceURL = "https://beta.edavki.durs.si/SoapPortal/EdpSoapService.asmx";

    private static int taxNumber = 12345679; // fake tax number, change it

    private static string taxPayerType = "FO"; // FO or PO or SP

    #endregion

    #region Public Methods and Operators

    /// <summary>
    /// The main entry point for the application.
    /// </summary>
    /// <param name="args">
    /// The args.
    /// </param>
    [STAThread]
    public static void Main(string[] args)
    {
      // Check arguments
      if (args.Length != 1 && args.Length != 6)
      {
        Console.Error.WriteLine(
          "usage: SoapPortalSample deposit|deposit_large|get "
          + "https://beta.edavki.durs.si/SoapPortal/EdpSoapService.asmx " + "beta.edavki.durs.si " + "c:\\user.cer "
          + "12345679 FO");
        Environment.Exit(1);
      }

      string action = args[0];

      // In case non-default parameters are used, parse these
      if (args.Length == 6)
      {
        soapServiceURL = args[1];
        documentDomain = args[2];
        X509Certificate certificateDer = X509Certificate.CreateFromCertFile(args[3]);
        certificateName = GetCNName(certificateDer);
        taxNumber = int.Parse(args[4]);
        taxPayerType = args[5];
      }

      // Load the certificate from the certificate store
      X509Store certificateStore = new X509Store(StoreName.My, StoreLocation.CurrentUser);
      certificateStore.Open(OpenFlags.ReadOnly);
      X509Certificate2Collection certificateCollection =
        certificateStore.Certificates.Find(X509FindType.FindBySubjectName, certificateName, true);

      X509Certificate2 userCertificate = certificateCollection[0];

      // Create SoapPortal client for specified URL and identify using client certificate
      EdpSoapServiceSoapClient soapService = new EdpSoapServiceSoapClient();
      soapService.ClientCredentials.ClientCertificate.Certificate = userCertificate;
      soapService.Endpoint.Address = new EndpointAddress(soapServiceURL);

      // Login, representing the chosen tax payer
      string sessionHandle = soapService.LoginUsingClientCertificate(taxNumber, taxPayerType);

      // Execute action
      switch (action)
      {
        case "deposit":
          ActionDeposit(soapService, userCertificate, sessionHandle);
          break;
        case "deposit_large":
          ActionDepositLarge(soapService, userCertificate, sessionHandle);
          break;
        case "get":
          ActionGet(soapService, sessionHandle);
          break;
      }

      // Logout
      soapService.Logout(sessionHandle);
    }

    #endregion

    #region Methods

    /// <summary>
    /// Create new NF_PrevZav document and deposit it.
    /// </summary>
    /// <param name="soapService">
    /// The SOAP service.
    /// </param>
    /// <param name="userCertificate">
    /// The user certificate.
    /// </param>
    /// <param name="sessionHandle">
    /// The session handle.
    /// </param>
    private static void ActionDeposit(
      EdpSoapServiceSoapClient soapService, X509Certificate2 userCertificate, string sessionHandle)
    {
      // Use SOAP service to create new document, then fill it with sample data
      string prevZavXml = CreateNfPrevZavDocument(soapService, sessionHandle);

      // Get the current display transform accepted by FURS
      string displayTransformHashSha1;
      string displayTransformUrl = soapService.GetDocumentDisplayTransform(
        sessionHandle, "NF_PrevZav", 1, out displayTransformHashSha1);

      // We should display document to user using the transformation before signing

      // Sign the document
      string prevZavXmlSigned = SignDepositDocument(
        prevZavXml, userCertificate, displayTransformUrl, displayTransformHashSha1);

      // Deposit the document
      string documentId = soapService.DepositDocument(sessionHandle, prevZavXmlSigned, false, null);

      Console.Out.WriteLine("Document Number = " + documentId);

      // Get the deposited document
      string depositedPrevZavXml;
      soapService.GetDepositedDocument(sessionHandle, documentId, out depositedPrevZavXml);
      Console.Out.WriteLine("Deposited document = " + depositedPrevZavXml);
    }

    /// <summary>
    /// Create new DMV-O document with internal attachment and deposit it. 
    /// Since the document is large, it has to be deposited using method DepositDocumentWithAttachment.
    /// </summary>
    /// <param name="soapService">
    /// The SOAP service.
    /// </param>
    /// <param name="userCertificate">
    /// The user certificate.
    /// </param>
    /// <param name="sessionHandle">
    /// The session handle.
    /// </param>
    private static void ActionDepositLarge(
      EdpSoapServiceSoapClient soapService, X509Certificate2 userCertificate, string sessionHandle)
    {
      // Create the document
      XmlDocument dmvoDoc = CreateDmvoDocument(soapService, sessionHandle);

      // Add internal attachment
      byte[] internalAttachmentData = Encoding.ASCII.GetBytes("I am an internal attachment.");

      string fileHash;
      int attachmentId = soapService.UploadInternalAttachment(
        sessionHandle, internalAttachmentData, "text/plain", "priloga1.txt", out fileHash);

      XmlNamespaceManager dmvoNsmgr = new XmlNamespaceManager(dmvoDoc.NameTable);
      dmvoNsmgr.AddNamespace("edp", EdpUri);
      dmvoNsmgr.AddNamespace("doc", DmvoUri);

      XmlNode attachmentListNode = dmvoDoc.SelectSingleNode("//doc:Envelope/edp:AttachmentList", dmvoNsmgr);
      XmlNode attachmentNode = CreateAndAppendChild(dmvoDoc, attachmentListNode, "ExternalAttachment", EdpUri);
      CreateAndAppendChild(dmvoDoc, attachmentNode, "attachmentId", EdpUri, attachmentId.ToString());
      CreateAndAppendChild(dmvoDoc, attachmentNode, "type", EdpUri, "Drugo");
      CreateAndAppendChild(dmvoDoc, attachmentNode, "filename", EdpUri, "priloga1.txt");
      CreateAndAppendChild(dmvoDoc, attachmentNode, "mimetype", EdpUri, "text/plain");
      CreateAndAppendChild(dmvoDoc, attachmentNode, "hash", EdpUri, fileHash);
      CreateAndAppendChild(dmvoDoc, attachmentNode, "description", EdpUri, "description");

      // Convert the document to byte array and calculate hash
      byte[] attachment = Encoding.UTF8.GetBytes(dmvoDoc.OuterXml);
      string attachmentHash = CalculateHashHex(attachment);

      // Transform the document using the large transformation for it
      string transformUrl = soapService.GetDocumentLargeTransform(sessionHandle, "DMV_O", 4);
      XslCompiledTransform xslTransform = new XslCompiledTransform();
      xslTransform.Load(transformUrl);

      XPathDocument largeDoc;
      using (MemoryStream ms = new MemoryStream(attachment))
      {
        largeDoc = new XPathDocument(ms);
      }

      XmlDocument transformedDoc = new XmlDocument { PreserveWhitespace = true };
      using (MemoryStream ms = new MemoryStream())
      {
        xslTransform.Transform(largeDoc, null, ms);
        ms.Position = 0;
        transformedDoc.Load(ms);
      }

      // Add the attachment hash to the transformed document
      XmlNamespaceManager nsmgr = new XmlNamespaceManager(transformedDoc.NameTable);
      nsmgr.AddNamespace("doc", DmvoUri);
      transformedDoc.SelectSingleNode("//doc:Attachment/doc:Hash", nsmgr).InnerText = attachmentHash;

      // Get the current display transform accepted by FURS
      string displayTransformHashSha1;
      string displayTransformUrl = soapService.GetDocumentDisplayTransform(
        sessionHandle, "DMV_O", 4, out displayTransformHashSha1);

      // Sign the document
      string dmvoSigned = SignDepositDocument(
        transformedDoc.OuterXml, userCertificate, displayTransformUrl, displayTransformHashSha1);

      // Deposit the document
      string documentId = soapService.DepositDocumentWithAttachment(
        sessionHandle, dmvoSigned, attachment, "DMV_O_4.xml", "text/xml", true, null);

      Console.Out.WriteLine("Document Number = " + documentId);

      // Get the deposited document
      string depositedDmvoXml;
      soapService.GetDepositedDocument(sessionHandle, documentId, out depositedDmvoXml);
      Console.Out.WriteLine("Deposited document = " + depositedDmvoXml);
    }

    /// <summary>
    /// Get the list of received documents.
    /// </summary>
    /// <param name="soapService">
    /// The SOAP service.
    /// </param>
    /// <param name="sessionHandle">
    /// The session handle.
    /// </param>
    private static void ActionGet(EdpSoapServiceSoapClient soapService, string sessionHandle)
    {
      string docList = soapService.GetReceivedDocumentList(sessionHandle, DateTime.Now.AddDays(-10), DateTime.Now);
      Console.Out.WriteLine("document list:\n" + docList);

      MemoryStream documentStream = new MemoryStream(Encoding.UTF8.GetBytes(docList));

      XmlDocument doc = new XmlDocument();
      doc.Load(new XmlTextReader(documentStream));

      XmlNamespaceManager nsmgr = new XmlNamespaceManager(doc.NameTable);
      nsmgr.AddNamespace("dl", DocumentInfoUri);

      XmlElement docNum = (XmlElement)doc.SelectSingleNode("//dl:DocumentInfo/dl:documentNumber", nsmgr);

      string docXml = null;
      if (docNum != null)
      {
        soapService.GetDepositedDocument(sessionHandle, docNum.InnerText, out docXml);
      }

      Console.Out.WriteLine("latest result document:\n" + docXml);
    }

    /// <summary>
    /// Fill actual data for document NF-PrevZav
    /// </summary>
    /// <param name="soapService">
    /// The SOAP service.
    /// </param>
    /// <param name="sessionHandle">
    /// The session handle.
    /// </param>
    /// <returns>
    /// The add payload nf prev zav.
    /// </returns>
    private static string CreateNfPrevZavDocument(EdpSoapServiceSoapClient soapService, string sessionHandle)
    {
      string prevZavXml = soapService.GetNewDocument(sessionHandle, "NF_PrevZav", 1);

      XmlDocument prevZavDoc = new XmlDocument { PreserveWhitespace = true };
      prevZavDoc.LoadXml(prevZavXml);

      XmlNamespaceManager nsmgr = new XmlNamespaceManager(prevZavDoc.NameTable);
      nsmgr.AddNamespace("doc", NfPrevZavUri);

      XmlNode nfPrevZavNode = prevZavDoc.SelectSingleNode("//doc:Envelope/doc:body/doc:NF_PrevZav", nsmgr);

      XmlNode verifyNode = CreateAndAppendChild(prevZavDoc, nfPrevZavNode, "Verify", NfPrevZavUri);
      CreateAndAppendChild(prevZavDoc, verifyNode, "TaxNumber", NfPrevZavUri, "48889679");
      CreateAndAppendChild(prevZavDoc, verifyNode, "Criteria", NfPrevZavUri);

      return prevZavDoc.OuterXml;
    }

    /// <summary>
    /// Calculates the SHA1 hash for the data.
    /// </summary>
    /// <param name="data">The document context.</param>
    /// <returns></returns>
    private static string CalculateHashHex(byte[] data)
    {
      SHA1 sha = new SHA1CryptoServiceProvider();
      byte[] arr = sha.ComputeHash(data);
      StringBuilder sb = new StringBuilder(arr.Length * 2);
      foreach (byte b in arr)
      {
        sb.Append(b.ToString("x2"));
      }

      return sb.ToString();
    }

    /// <summary>
    /// Creates the child element with the specified local name and namespace and appends it to the specified parent.
    /// </summary>
    /// <param name="xmlDoc">The XML doc.</param>
    /// <param name="parent">The parent.</param>
    /// <param name="localName">Name of the local.</param>
    /// <param name="namespaceUri">The namespace URI.</param>
    /// <param name="innerText">The inner text.</param>
    /// <returns>The created child element.</returns>
    private static XmlNode CreateAndAppendChild(
      XmlDocument xmlDoc, XmlNode parent, string localName, string namespaceUri, string innerText = null)
    {
      XmlNode child = xmlDoc.CreateElement(string.Empty, localName, namespaceUri);
      parent.AppendChild(child);
      if (innerText != null)
      {
        child.InnerText = innerText;
      }

      return child;
    }

    /// <summary>
    /// Creates a new DMV-O document and fills it with sample data
    /// </summary>
    /// <param name="soapService">
    /// The SOAP service.
    /// </param>
    /// <param name="sessionHandle">
    /// The session handle.
    /// </param>
    /// <returns>
    /// The created document.
    /// </returns>
    private static XmlDocument CreateDmvoDocument(EdpSoapServiceSoapClient soapService, string sessionHandle)
    {
      // Since the SOAP portal does not support creating new DMV-O document, we copy the tax payer 
      // data from ODE_PDO document, which can be created.
      string odePdoXml = soapService.GetNewDocumentForDate(sessionHandle, "ODE_PDO", new DateTime(2015, 1, 1));

      XmlDocument prevZavDoc = new XmlDocument();
      prevZavDoc.LoadXml(odePdoXml);

      XmlNamespaceManager odePdoNsmgr = new XmlNamespaceManager(prevZavDoc.NameTable);
      odePdoNsmgr.AddNamespace("edp", EdpUri);

      XmlNode prevZavTaxpayerNode = prevZavDoc.SelectSingleNode("//edp:Header/edp:taxpayer", odePdoNsmgr);

      XmlDocument dmvoDoc = new XmlDocument();
      dmvoDoc.LoadXml(DmvoTemplate);

      XmlNamespaceManager dmvoNsmgr = new XmlNamespaceManager(dmvoDoc.NameTable);
      dmvoNsmgr.AddNamespace("edp", EdpUri);
      dmvoNsmgr.AddNamespace("doc", DmvoUri);

      XmlNode taxpayerNode = dmvoDoc.ImportNode(prevZavTaxpayerNode, true);
      XmlNode headerNode = dmvoDoc.SelectSingleNode("//edp:Header", dmvoNsmgr);
      headerNode.AppendChild(taxpayerNode);

      // Fill large data
      XmlNode dmvoNode = dmvoDoc.SelectSingleNode("//doc:Envelope/doc:body/doc:DMV_O", dmvoNsmgr);

      for (int i = 1; i <= 105; i++)
      {
        XmlNode vehicleNode = CreateAndAppendChild(dmvoDoc, dmvoNode, "Vehicle", DmvoUri);
        CreateAndAppendChild(dmvoDoc, vehicleNode, "SequenceNumber", DmvoUri, i.ToString());
        CreateAndAppendChild(dmvoDoc, vehicleNode, "VehicleType", DmvoUri, "A");
        CreateAndAppendChild(dmvoDoc, vehicleNode, "VehicleMake", DmvoUri, "makeX");
        CreateAndAppendChild(dmvoDoc, vehicleNode, "VehicleModel", DmvoUri, "modelY");
        CreateAndAppendChild(dmvoDoc, vehicleNode, "CodeVIN", DmvoUri, "1000A000000000" + i.ToString("000"));
        CreateAndAppendChild(dmvoDoc, vehicleNode, "FuelTypeCode", DmvoUri, "D");
        CreateAndAppendChild(dmvoDoc, vehicleNode, "FuelTypeName", DmvoUri, "dizel");
        CreateAndAppendChild(dmvoDoc, vehicleNode, "CO2Omission", DmvoUri, "1");
        CreateAndAppendChild(dmvoDoc, vehicleNode, "EnginePower", DmvoUri, "100.00");
        CreateAndAppendChild(dmvoDoc, vehicleNode, "EngineVolume", DmvoUri, "3333.3");
        CreateAndAppendChild(dmvoDoc, vehicleNode, "HeavyPartsOmission", DmvoUri, "3.00000");
        CreateAndAppendChild(dmvoDoc, vehicleNode, "EuroOmissionRateCode", DmvoUri, "EURO1");
        CreateAndAppendChild(dmvoDoc, vehicleNode, "EuroOmissionRateName", DmvoUri, "EURO1");
        CreateAndAppendChild(dmvoDoc, vehicleNode, "SeatCount", DmvoUri, "5");
        CreateAndAppendChild(dmvoDoc, vehicleNode, "CertificateOfConformityIssuingDate", DmvoUri, "2012-01-01");
        CreateAndAppendChild(dmvoDoc, vehicleNode, "CertificateOfConformity", DmvoUri, "true");
        CreateAndAppendChild(dmvoDoc, vehicleNode, "SupplyEUCountry", DmvoUri, "AT");
        CreateAndAppendChild(dmvoDoc, vehicleNode, "SupplyEUCountryName", DmvoUri, "AUSTRIA");
        CreateAndAppendChild(dmvoDoc, vehicleNode, "TaxBase", DmvoUri, "1000.00");
        CreateAndAppendChild(dmvoDoc, vehicleNode, "TaxRate", DmvoUri, "16.00");
        CreateAndAppendChild(dmvoDoc, vehicleNode, "TaxAmount", DmvoUri, "160.00");
        CreateAndAppendChild(dmvoDoc, vehicleNode, "TaxAddRate", DmvoUri, "10.00");
        CreateAndAppendChild(dmvoDoc, vehicleNode, "TaxAddAmount", DmvoUri, "100.00");
        CreateAndAppendChild(dmvoDoc, vehicleNode, "TaxSum", DmvoUri, "260.00");
      }

      return dmvoDoc;
    }

    /// <summary>
    /// Get Name from Certificate CN (remove + OID... Serial Number)
    /// </summary>
    /// <param name="cert">
    /// certificate
    /// </param>
    /// <returns>
    /// name from subject's CN
    /// </returns>
    private static string GetCNName(X509Certificate cert)
    {
      string subject = cert.Subject;

      string cnname = subject;
      int index = cnname.IndexOf("CN=");

      if (index != -1)
      {
        cnname = subject.Substring(index + 3);

        char[] separator = new char[2];
        separator[0] = ',';
        separator[1] = '+';

        string[] fields = cnname.Split(separator);

        cnname = fields[0];
      }

      return cnname.Trim();
    }

    private static string PrepareDepositDocument(
      string document, string displayTransformURL, string displayTransformHashSHA1, X509Certificate certificate)
    {
      MemoryStream documentStream = new MemoryStream(Encoding.UTF8.GetBytes(document));

      XmlDocument doc = new XmlDocument();

      // doc.PreserveWhitespace = true;
      doc.Load(new XmlTextReader(documentStream));

      XmlNamespaceManager nsmgr = new XmlNamespaceManager(doc.NameTable);
      nsmgr.AddNamespace("ds", DsigUri);
      nsmgr.AddNamespace("edp", EdpUri);
      nsmgr.AddNamespace("hslsig", HslsigUri);
      XmlNode sigs = doc.SelectSingleNode("//edp:Signatures", nsmgr);

      XmlElement depositorSignature = doc.CreateElement("edp:DepositorSignature", EdpUri);
      sigs.AppendChild(depositorSignature);

      XmlElement depositor = doc.CreateElement("edp:Depositor", EdpUri);
      depositorSignature.AppendChild(depositor);

      XmlElement timestamp = doc.CreateElement("edp:timestamp", EdpUri);
      depositor.AppendChild(timestamp);

      XmlElement name = doc.CreateElement("edp:name", EdpUri);
      depositor.AppendChild(name);

      name.InnerText = GetCNName(certificate);
      timestamp.InnerText = XmlConvert.ToString(DateTime.Now, XmlDateTimeSerializationMode.Local);

      XmlElement presentation = doc.CreateElement("hslsig:Presentation", HslsigUri);
      depositorSignature.AppendChild(presentation);

      string transformUri = displayTransformURL;
      string transformDigestSha1 = displayTransformHashSHA1;

      XmlElement displayTransformDigest = doc.CreateElement("hslsig:DisplayTransformDigest", HslsigUri);
      displayTransformDigest.SetAttribute("transform", transformUri);
      displayTransformDigest.SetAttribute("digestAlgorithm", "http://www.w3.org/2000/09/xmldsig#sha1");
      displayTransformDigest.SetAttribute("digest", transformDigestSha1);
      presentation.AppendChild(displayTransformDigest);

      XmlElement signingComponent = doc.CreateElement("hslsig:SigningComponent", HslsigUri);
      signingComponent.SetAttribute("name", "SoapPortalSample");
      signingComponent.SetAttribute("version", "0");
      presentation.AppendChild(signingComponent);

      XmlElement signature = doc.CreateElement("ds:Signature", DsigUri);
      signature.SetAttribute("Id", "DepositorSignature");
      depositorSignature.AppendChild(signature);

      XmlNode header = doc.SelectSingleNode("//edp:Header", nsmgr);
      XmlNode domain = header.SelectSingleNode("//edp:Header/edp:domain", nsmgr);
      if (domain == null)
      {
        domain = doc.CreateElement("edp:domain", EdpUri);
        header.AppendChild(domain);
      }

      domain.InnerXml = documentDomain;

      return doc.OuterXml;
    }

    private static string SignDepositDocument(
      string document, X509Certificate2 signCertificate, string displayTransformURL, string displayTransformHashSHA1)
    {
      string toSign = PrepareDepositDocument(document, displayTransformURL, displayTransformHashSHA1, signCertificate);
      string signed = SignDocument(toSign, "DepositorSignature", signCertificate);
      return signed;
    }

    private static string SignDocument(string document, string signatureID, X509Certificate2 signCertificate)
    {
      // Obtain private key
      AsymmetricAlgorithm rsa = signCertificate.PrivateKey;

      // Read provided document, find the signature element ...
      MemoryStream documentStream = new MemoryStream(System.Text.Encoding.UTF8.GetBytes(document));

      XmlDocument doc = new XmlDocument();
      doc.PreserveWhitespace = true;
      doc.Load(new XmlTextReader(documentStream));

      XmlNamespaceManager nsmgr = new XmlNamespaceManager(doc.NameTable);
      nsmgr.AddNamespace("ds", DsigUri);
      XmlNode sig = doc.SelectSingleNode("//*[@Id='" + signatureID + "']", nsmgr);

      XmlNode after = sig.PreviousSibling;
      XmlNode parent = sig.ParentNode;

      SignedXml signedXml = new SignedXml(doc);
      parent.RemoveChild(sig);

      Reference r = new Reference();
      r.Uri = string.Empty;
      r.AddTransform(new XmlDsigEnvelopedSignatureTransform());
      signedXml.AddReference(r);

      X509Certificate cert1 = new X509Certificate(signCertificate.GetRawCertData());

      KeyInfo ki = new KeyInfo();
      ki.AddClause(new KeyInfoX509Data(cert1));
      signedXml.KeyInfo = ki;

      signedXml.Signature.Id = signatureID;

      signedXml.SigningKey = rsa;
      signedXml.ComputeSignature();

      XmlNode signature = doc.ImportNode(signedXml.GetXml(), true);
      parent.InsertAfter(signature, after);

      return doc.OuterXml;
    }

    #endregion
  }
}